# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""

Class for storing macauth nonces in memcached.

"""

import time
import math
from hashlib import sha1
from base64 import urlsafe_b64encode

from mozsvc.storage.mcclient import MemcachedClient


DEFAULT_NONCE_TTL = 30  # 30 seconds
DEFAULT_ID_TTL = 3660   # 1 hour


class MemcachedNonceCache(object):
    """Object for managing a cache of used nonce values in memcached.

    This class allow easy timestamp-based management of client-generated
    nonces according to the rules of RFC-TODO:

        * Maintain a measure of clock skew for each MAC id.
        * Reject nonces with a timestamp outside the configured range.
        * Reject nonces that have already been seen.

    It stores the nonces in memcached so that they can be shared between
    different webserver processes.  The clock-skew for each id token is
    stored under they key sha1(<id>:skew), while each nonce seen for the id
    is marked by a key sha1(<id>:nonce:<timestamp>:<nonce>).

    NOTE: the "MAC id" here corresponds to the full authentication token
    issues by the tokenserver, not to the numeric userid of an individual
    user.  So it is entirely possible to have multiple skew records for
    each user, corresponding to different active tokens.
    """

    def __init__(self, nonce_ttl=None, id_ttl=None, cache_servers=None,
                 cache_key_prefix="noncecache:", cache_pool_size=None,
                 cache_pool_timeout=60, **kwds):
        # Memcached ttls are in integer seconds, so round up to the nearest.
        if nonce_ttl is None:
            nonce_ttl = DEFAULT_NONCE_TTL
        else:
            nonce_ttl = int(math.ceil(nonce_ttl))
        if id_ttl is None:
            id_ttl = DEFAULT_ID_TTL
        else:
            id_ttl = int(math.ceil(id_ttl))
        self.nonce_ttl = nonce_ttl
        self.id_ttl = id_ttl
        self.mcclient = MemcachedClient(cache_servers, cache_key_prefix,
                                        cache_pool_size, cache_pool_timeout)

    def _key(self, *names):
        """Get a memcached key built from the given component names.

        This method returns the memcached key to use for the given component
        names, by contatentating them together and then hashing them.  The
        hashing serves both to ensure confidentiality of the macauth tokens
        store in memcached, and the reduce the size of the keys.
        """
        return urlsafe_b64encode(sha1(":".join(names)).digest())

    def check_nonce(self, id, timestamp, nonce):
        """Check if the given timestamp+nonce is fresh for the given id.

        This method checks that the given timestamp+nonce has not previously
        been seen for the given id.  It returns True if the nonce is fresh
        and False if not.

        Fresh nonces are added to the cache, so that subsequent checks of the
        same nonce will return False.
        """
        # We want to fetch the recorded clock skew for this id, along with
        # any existing cache entry for the provided nonce.
        key_skew = self._key(id, "skew")
        key_nonce = self._key(id, "nonce", str(timestamp), nonce)
        # Use get_multi to fetch both keys in a single request.
        # If the data appears to be corrupted then fail out for safety.
        try:
            cached = self.mcclient.get_multi([key_skew, key_nonce])
        except ValueError:
            return False
        # If the nonce appears in the cache, it must be stale.
        if key_nonce in cached:
            return False
        # If we've never recorded a clock skew for this id, record it now.
        try:
            skew = cached[key_skew]
        except KeyError:
            skew = int(time.time() - timestamp)
            self.mcclient.add(key_skew, skew, time=self.id_ttl)
        # If the adjusted timestamp is too old or too new, it is stale.
        # XXX TODO: we should use a monotonic clock here.
        if abs(timestamp + skew - time.time()) >= self.nonce_ttl:
            return False
        # The nonce is fresh, add it into the cache.
        self.mcclient.add(key_nonce, True, time=self.nonce_ttl)
        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from webob import Response
import urllib2
import socket
import base64
from urlparse import urlparse, urlunparse


def get_url(url, method='GET', data=None, user=None, password=None, timeout=5,
            get_body=True, extra_headers=None):
    """Performs a synchronous url call and returns the status and body.

    This function is to be used to provide a gateway service.

    If the url is not answering after `timeout` seconds, the function will
    return a (504, {}, error).

    If the url is not reachable at all, the function will
    return (502, {}, error)

    Other errors are managed by the urrlib2.urllopen call.

    Args:
        - url: url to visit
        - method: method to use
        - data: data to send
        - user: user to use for Basic Auth, if needed
        - password: password to use for Basic Auth
        - timeout: timeout in seconds.
        - extra headers: mapping of headers to add
        - get_body: if set to False, the body is not retrieved

    Returns:
        - tuple : status code, headers, body
    """
    if isinstance(password, unicode):
        password = password.encode('utf-8')

    req = urllib2.Request(url, data=data)
    req.get_method = lambda: method

    if user is not None and password is not None:
        auth = base64.encodestring('%s:%s' % (user, password))
        req.add_header("Authorization", "Basic %s" % auth.strip())

    if extra_headers is not None:
        for name, value in extra_headers.items():
            req.add_header(name, value)

    try:
        res = urllib2.urlopen(req, timeout=timeout)
    except urllib2.HTTPError, e:
        if hasattr(e, 'headers'):
            headers = dict(e.headers)
        else:
            headers = {}

        if hasattr(e, 'read'):
            body = e.read()
        else:
            body = ''

        return e.code, headers, body

    except urllib2.URLError, e:
        if isinstance(e.reason, socket.timeout):
            return 504, {}, str(e)
        return 502, {}, str(e)

    if get_body:
        body = res.read()
    else:
        body = ''

    return res.getcode(), dict(res.headers), body


def proxy(request, scheme, netloc, timeout=5):
    """Proxies and return the result from the other server.

    - scheme: http or https
    - netloc: proxy location
    """
    parsed = urlparse(request.url)
    path = parsed.path
    params = parsed.params
    query = parsed.query
    fragment = parsed.fragment
    url = urlunparse((scheme, netloc, path, params, query, fragment))
    method = request.method
    data = request.body

    # copying all X- headers
    xheaders = {}
    for header, value in request.headers.items():
        if not header.startswith('X-'):
            continue
        xheaders[header] = value

    if 'X-Forwarded-For' not in request.headers:
        xheaders['X-Forwarded-For'] = request.remote_addr

    if hasattr(request, '_authorization'):
        xheaders['Authorization'] = request._authorization

    status, headers, body = get_url(url, method, data, timeout=timeout,
                                    extra_headers=xheaders)

    return Response(body, status, headers.items())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
from StringIO import StringIO
from textwrap import dedent
import functools
import unittest2

from pyramid import testing
from pyramid.config import Configurator
from webtest import TestApp

from cornice.tests import CatchErrors
from mozsvc.config import Config
metlog = True
try:
    from mozsvc.metrics import MetricsService
    from mozsvc.plugin import load_from_config
except ImportError:
    metlog = False
    from cornice import Service as MetricsService  # NOQA

service3 = MetricsService(name="service3", path="/service3")
service4 = MetricsService(name="service4", path="/service4")
service5 = MetricsService(name="service5", path="/service5")


def wrap_fn(fn):
    if not hasattr(fn, '_wrap_count'):
        fn._wrap_count = 0
    else:
        fn._wrap_count += 1

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        result = fn(*args, **kwargs)
        result["wrapped%d" % fn._wrap_count] = "yes"
        return result
    return wrapper


@service3.get(decorators=[wrap_fn])
def wrapped_get3(request):
    return {"test": "succeeded"}


@service4.post(decorators=[wrap_fn])
@service4.get(decorators=[wrap_fn])
def wrapped_get4(request):
    return {"test": "succeeded"}


@service5.get(decorators=[wrap_fn])
@service5.get(accept="application/json", renderer="simplejson")
@service5.get(accept="application/newlines", renderer="newlines")
@service5.post(decorators=[wrap_fn])
def wrapped_get5(request):
    return {"test": "succeeded"}


class TestServiceDefinition(unittest2.TestCase):

    def setUp(self):
        if not metlog:
            raise(unittest2.SkipTest('no metlog'))
        mozconfig = Config(StringIO(dedent("""
        [test1]
        backend = mozsvc.metrics.MetlogPlugin
        sender_class=metlog.senders.DebugCaptureSender
        """)))
        settings = {"config": mozconfig}
        self.plugin = load_from_config("test1", mozconfig)
        self.config = Configurator(settings=settings)
        self.config.include("cornice")
        self.config.scan("mozsvc.tests.test_service_definition")
        self.app = TestApp(CatchErrors(self.config.make_wsgi_app()))

    def tearDown(self):
        testing.tearDown()

    def test_decorated_view_fn(self):
        # passing a decorator in to the service api call should result in a
        # decorated view callable
        resp = self.app.get("/service3")
        self.assertEquals(resp.json, {'test': 'succeeded', 'wrapped0': 'yes'})

    def test_stacked_decorated_view(self):
        # passing a decorator in to the service api call should result in a
        # decorated view callable, ordering of the particular decorators
        # shouldn't break things
        resp = self.app.get("/service4")
        self.assertEquals(resp.json, {'test': 'succeeded', 'wrapped0': 'yes'})

        resp = self.app.get("/service5")
        self.assertEquals(resp.json, {'test': 'succeeded', 'wrapped0': 'yes'})



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
"""

Utilities for authentication via the Sagrada auth system.

"""

from zope.interface import implements

from pyramid.request import Request
from pyramid.security import authenticated_userid
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.httpexceptions import HTTPUnauthorized

from pyramid_macauth import MACAuthenticationPolicy

import tokenlib

from cef import log_cef, AUTH_FAILURE

import mozsvc
import mozsvc.secrets

import logging
logger = logging.getLogger("mozsvc.user")


ENVIRON_KEY_IDENTITY = "mozsvc.user.identity"


class RequestWithUser(Request):
    """Request object that exposes the current user as "request.user".

    This is a custom Pyramid Request class that does user authentication
    on demand.  When you access the "user" attribute it will automatically
    search for credentials in the request, authenticate them, create a dict
    of user data and return it.

    If there are no credentials in the request then the "user" attribute will
    be an empty dict.  If the call to authenticated_userid() raises an error
    it will *not* be intercepted, you will receive it as an error from the
    attribute access attempt.
    """

    def _get_user(self):
        # Return an existing identity if there is one.
        user = self.environ.get(ENVIRON_KEY_IDENTITY)
        if user is not None:
            return user
        # Put an empty identity dict in place before calling the authn policy.
        # This lets the policy store stuff in req.user.
        user = self.environ[ENVIRON_KEY_IDENTITY] = {}
        # Call the standard authn framework to authenticate.
        try:
            userid = authenticated_userid(self)
        except Exception:
            self.environ.pop(ENVIRON_KEY_IDENTITY, None)
            raise
        # If the auth was successful, store the userid in the identity.
        if userid is not None:
            user.setdefault("uid", userid)
        return user

    def _set_user(self, user):
        self.environ["mozsvc.user.identity"] = user

    user = property(_get_user, _set_user)


class SagradaAuthenticationPolicy(MACAuthenticationPolicy):
    """Pyramid authentication policy for use with Sagrada auth tokens.

    This class provides an IAuthenticationPolicy implementation based on
    Sagrada authentication tokens as described here:

        https://wiki.mozilla.org/Services/Sagrada/TokenServer

    For verification of token signatures, this plugin can use either a
    single fixed secret (via the argument 'secret') or a file mapping
    node hostnames to secrets (via the argument 'secrets_file').  The
    two arguments are mutually exclusive.
    """

    implements(IAuthenticationPolicy)

    def __init__(self, secret=None, secrets_file=None, **kwds):
        if secret is not None and secrets_file is not None:
            msg = "Can only specify one of 'secret' or 'secrets_file'"
            raise ValueError(msg)
        elif secret is None and secrets_file is None:
            # Using secret=None will cause tokenlib to use a randomly-generated
            # secret.  This is useful for getting started without having to
            # twiddle any configuration files, but probably not what anyone
            # wants to use long-term.
            msgs = ["WARNING: using a randomly-generated token secret.",
                    "You probably want to set 'secret' or 'secrets_file' in"
                    "the [macauth] section of your configuration"]
            for msg in msgs:
                mozsvc.logger.warn(msg)
        if secrets_file is not None:
            self.secret = None
            self.secrets = mozsvc.secrets.Secrets(secrets_file)
        else:
            self.secret = secret
            self.secrets = None
        super(SagradaAuthenticationPolicy, self).__init__(**kwds)

    @classmethod
    def _parse_settings(cls, settings):
        """Parse settings for an instance of this class."""
        supercls = super(SagradaAuthenticationPolicy, cls)
        kwds = supercls._parse_settings(settings)
        for setting in ("secret", "secrets_file"):
            if setting in settings:
                kwds[setting] = settings.pop(setting)
        return kwds

    def _check_signature(self, request, key):
        """Check the MACAuth signature on the request.

        This method checks the MAC signature on the request against the
        supplied signing key.  If missing or invalid then HTTPUnauthorized
        is raised.

        The SagradaAuthenticationPolicy implementation wraps the default
        MACAuthenticationPolicy implementation with some cef logging.
        """
        supercls = super(SagradaAuthenticationPolicy, self)
        try:
            return supercls._check_signature(request, key)
        except HTTPUnauthorized:
            log_cef("Authentication Failed: invalid MAC signature", 5,
                    request.environ, request.registry.settings,
                    "", signature=AUTH_FAILURE)
            raise

    def decode_mac_id(self, request, tokenid):
        """Decode a MACAuth token id into its userid and MAC secret key.

        This method determines the appropriate secrets to use for the given
        request, then passes them on to tokenlib to handle the given MAC id
        token.

        If the id is invalid then ValueError will be raised.
        """
        # There might be multiple secrets in use, if we're in the
        # process of transitioning from one to another.  Try each
        # until we find one that works.
        secrets = self._get_token_secrets(request)
        for secret in secrets:
            try:
                data = tokenlib.parse_token(tokenid, secret=secret)
                userid = data["uid"]
                key = tokenlib.get_token_secret(tokenid, secret=secret)
                break
            except (ValueError, KeyError):
                pass
        else:
            log_cef("Authentication Failed: invalid MAC id", 5,
                    request.environ, request.registry.settings,
                    "", signature=AUTH_FAILURE)
            raise ValueError("invalid MAC id")
        return userid, key

    def encode_mac_id(self, request, userid):
        """Encode the given userid into a MAC id and secret key.

        This method is essentially the reverse of decode_mac_id.  It is
        not needed for consuming authentication tokens, but is very useful
        when building them for testing purposes.
        """
        # There might be multiple secrets in use, if we're in the
        # process of transitioning from one to another.  Always use
        # the last one aka the "most recent" secret.
        secret = self._get_token_secrets(request)[-1]
        tokenid = tokenlib.make_token({"uid": userid}, secret=secret)
        key = tokenlib.get_token_secret(tokenid, secret=secret)
        return tokenid, key

    def _get_token_secrets(self, request):
        """Get the list of possible secrets for signing tokens."""
        if self.secrets is None:
            return [self.secret]
        # Secrets are looked up by hostname.
        # We need to normalize some port information for this work right.
        node_name = request.host_url
        if node_name.startswith("http:") and node_name.endswith(":80"):
            node_name = node_name[:-3]
        elif node_name.startswith("https:") and node_name.endswith(":443"):
            node_name = node_name[:-4]
        return self.secrets.get(node_name)


def includeme(config):
    """Include mozsvc user-handling into a pyramid config.

    This function will set up user-handling via the mozsvc.user system.
    Things configured include:

        * use RequestWithUser as the request object factory
        * use SagradaAuthenticationPolicy as the default authn policy

    """
    # Use RequestWithUser as the request object factory.
    config.set_request_factory(RequestWithUser)

    # Hook up a default AuthorizationPolicy.
    # ACLAuthorizationPolicy is usually what you want.
    # If the app configures one explicitly then this will get overridden.
    # In auto-commit mode this needs to be set before adding an authn policy.
    authz_policy = ACLAuthorizationPolicy()
    config.set_authorization_policy(authz_policy)

    # Build a SagradaAuthenticationPolicy from the deployment settings.
    settings = config.get_settings()
    authn_policy = SagradaAuthenticationPolicy.from_settings(settings)
    config.set_authentication_policy(authn_policy)

    # Set the forbidden view to use the challenge() method from the policy.
    config.add_forbidden_view(authn_policy.challenge)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest

import pyramid.testing

from mozsvc.exceptions import BackendError
from mozsvc.tests.support import make_request


class TestTweens(unittest.TestCase):

    def setUp(self):
        self.app = None
        self.config = pyramid.testing.setUp()
        self.config.registry.settings["mozsvc.retry_after"] = "17"
        self.config.include("mozsvc")
        self.config.add_route("backend_error", "/backend_error")

    def tearDown(self):
        pyramid.testing.tearDown()

    def _make_request(self, *args, **kwds):
        return make_request(self.config, *args, **kwds)

    def _do_request(self, *args, **kwds):
        if self.app is None:
            self.app = self.config.make_wsgi_app()
        req = self._make_request(*args, **kwds)
        return self.app.handle_request(req)

    def _set_backend_error_view(self, func):
        self.config.add_view(func, route_name="backend_error")

    def test_that_backend_errors_are_captured(self):
        @self._set_backend_error_view
        def backend_error(request):
            raise BackendError

        r = self._do_request("/backend_error")
        self.assertEquals(r.status_int, 503)
        self.assertEquals(r.headers["Retry-After"], "17")

    def test_that_backend_errors_can_set_retry_after(self):
        @self._set_backend_error_view
        def backend_error(request):
            raise BackendError(retry_after=42)

        r = self._do_request("/backend_error")
        self.assertEquals(r.status_int, 503)
        self.assertEquals(r.headers["Retry-After"], "42")

    def test_that_retry_after_doesnt_get_set_to_zero(self):
        @self._set_backend_error_view
        def backend_error(request):
            raise BackendError(retry_after=0)

        r = self._do_request("/backend_error")
        self.assertEquals(r.status_int, 503)
        self.assertEquals(r.headers.get("Retry-After"), None)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
""" Exceptions
"""
from ConfigParser import Error


class BackendError(Exception):
    """Raised when the backend is down or fails"""
    def __init__(self, msg='', server='', retry_after=None,
                 backend=None, request=None):
        """
        - msg, server will be dumped in str() if provided
        - retry_after, if set to a positive integer, will be used to send
          back a Retry-After header value. If not set, a default value is
          returned. If set to 0, the header is explicitely skipped.
        - backend: if provided, the backend from where originated the error.
          it will be used to render more details, It can be any kind of object
          as __str__() will be called on it.

        """
        self.msg = msg
        self.server = server
        self.retry_after = retry_after
        self.backend = backend
        self.request = request

    def __str__(self):
        res = self.__class__.__name__
        if self.server != '':
            res += ' on %s' % self.server
        if self.backend is not None:
            res += '\n%s' % str(self.backend)
        if self.msg != '':
            res += '\n\n%s' % self.msg
        if self.request:
            res = '%s %s\n%s' % (self.request.method,
                                 self.request.path_info,
                                 res)
        return res


class BackendTimeoutError(BackendError):
    """Raised when the backend times out."""
    pass


class MaxConnectionReachedError(Exception):
    """Raised by ldappool"""
    pass


class NoEmailError(Exception):
    """Raised when we need the user's email address and it doesn't exist."""
    pass


class NoUserIDError(Exception):
    """Raised when there's no userID fails."""
    pass


class NodeAttributionError(Exception):
    """Raised when the node attribution fails."""
    pass


class InvalidCodeError(Exception):
    """Raised when we need the user's reset code is incorrect."""
    pass


class EnvironmentNotFoundError(Error):
    """Raised when an environment variable is not found"""

    def __init__(self, varname):
        Error.__init__(self, 'Variable not found %r' % varname)
        self.varname = varname


class MethodNotFoundError(Error):
    ''' Raised when a method lookup fails '''
    pass


""" 400 error codes

Warning

  If you add a constant here, please update
  https://hg.mozilla.org/services/docs/file/tip/source/respcodes.rst
  which is used to generate http://docs.services.mozilla.com/respcodes.html
"""

ERROR_ILLEGAL_METHOD = 1            # Illegal method/protocol
ERROR_INVALID_CAPTCHA = 2           # Incorrect/missing captcha
ERROR_INVALID_USER = 3              # Invalid/missing username
ERROR_INVALID_WRITE = 4             # Attempt to overwrite data that can't be
ERROR_WRONG_USERID = 5              # Userid does not match account in path
ERROR_MALFORMED_JSON = 6            # Json parse failure
ERROR_MISSING_PASSWORD = 7          # Missing password field
ERROR_INVALID_OBJECT = 8            # Invalid Basic Object
ERROR_WEAK_PASSWORD = 9             # Requested password not strong enough
ERROR_INVALID_RESET_CODE = 10       # Invalid/missing password reset code
ERROR_UNSUPPORTED_FUNCTION = 11     # Unsupported function
ERROR_NO_EMAIL_ADDRESS = 12         # No email address on file
ERROR_INVALID_COLLECTION = 13       # Invalid collection
ERROR_OVER_QUOTA = 14               # User over quota
ERROR_USERNAME_EMAIL_MISMATCH = 15  # The e-mail does not match the username



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest
import urllib2
import socket
from mozsvc.http_helpers import get_url, proxy


class FakeResult(object):
    headers = {}
    body = '{}'

    def getcode(self):
        return 200

    def read(self):
        return self.body


class TestHttp(unittest.TestCase):

    def setUp(self):
        self.oldopen = urllib2.urlopen
        urllib2.urlopen = self._urlopen

    def tearDown(self):
        urllib2.urlopen = self.oldopen

    def _urlopen(self, req, timeout=None):
        url = req.get_full_url()
        if url == 'impossible url':
            raise ValueError()
        if url == 'http://dwqkndwqpihqdw.com':
            msg = 'Name or service not known'
            raise urllib2.URLError(socket.gaierror(-2, msg))

        if url in ('http://google.com', 'http://goodauth'):
            return FakeResult()
        if url == 'http://badauth':
            raise urllib2.HTTPError(url, 401, '', {}, None)
        if url == 'http://timeout':
            raise urllib2.URLError(socket.timeout())
        if url == 'http://error':
            raise urllib2.HTTPError(url, 500, 'Error', {}, None)
        if url == 'http://newplace':
            res = FakeResult()
            res.body = url + ' ' + req.headers['Authorization']
            return res
        if url == 'http://xheaders':
            res = FakeResult()
            headers = req.headers.items()
            headers.sort()
            res.body = str(headers)
            return res

        raise ValueError(url)

    def test_get_url(self):

        # malformed url
        self.assertRaises(ValueError, get_url, 'impossible url')

        # unknown location
        code, headers, body = get_url('http://dwqkndwqpihqdw.com',
                                      get_body=False)
        self.assertEquals(code, 502)
        self.assertTrue('Name or service not known' in body)

        # any page
        code, headers, body = get_url('http://google.com', get_body=False)
        self.assertEquals(code, 200)
        self.assertEquals(body, '')

        # page with auth failure
        code, headers, body = get_url('http://badauth',
                                      user='tarek',
                                      password='xxxx')
        self.assertEquals(code, 401)

        # page with right auth
        code, headers, body = get_url('http://goodauth',
                                      user='tarek',
                                      password='passat76')
        self.assertEquals(code, 200)
        self.assertEquals(body, '{}')

        # page that times out
        code, headers, body = get_url('http://timeout', timeout=0.1)
        self.assertEquals(code, 504)

        # page that fails
        code, headers, body = get_url('http://error', get_body=False)
        self.assertEquals(code, 500)

    def test_proxy(self):
        class FakeRequest(object):
            url = 'http://locahost'
            method = 'GET'
            body = 'xxx'
            headers = {'Content-Length': 3, 'X-Me-This': 1,
                       'X-Me-That': 2}
            remote_addr = '192.168.1.1'
            _authorization = 'Basic SomeToken'

        request = FakeRequest()
        response = proxy(request, 'http', 'newplace')
        self.assertEqual(response.content_length, 31)
        self.assertEqual(response.body, 'http://newplace Basic SomeToken')

        # we want to make sure that X- headers are proxied
        request = FakeRequest()
        response = proxy(request, 'http', 'xheaders')
        self.assertTrue("('X-me-that', 2), ('X-me-this', 1)" in response.body)
        self.assertTrue("X-forwarded-for" in response.body)



from StringIO import StringIO
from mozsvc.config import Config, get_configurator
from mozsvc.plugin import load_from_config
from pyramid.config import Configurator
from textwrap import dedent
from webob.request import Request
import unittest2
import json

metlog = True
try:
    from metlog.client import MetlogClient
    from metlog.senders import ZmqPubSender
    from metlog.senders.logging import StdLibLoggingSender
    from metlog.decorators import incr_count
    from metlog.decorators import timeit
    from mozsvc.metrics import MetlogPlugin
    from mozsvc.metrics import MetricsService
    from mozsvc.metrics import load_metlog_client
    from mozsvc.metrics import send_mozsvc_data
    from mozsvc.metrics import update_mozsvc_data
except ImportError:
    metlog = False
    from cornice import Service as MetricsService  # NOQA
    timeit = send_mozsvc_data = incr_count = lambda fn: fn  # NOQA


class TestMetrics(unittest2.TestCase):
    def setUp(self):
        if not metlog:
            raise(unittest2.SkipTest('no metlog'))

    def test_loading_from_config(self):
        mozconfig = Config(StringIO(dedent("""
        [test1]
        backend = mozsvc.metrics.MetlogPlugin
        sender_class=metlog.senders.ZmqPubSender
        sender_bindstrs=tcp://localhost:5585
                        tcp://localhost:5586

        [test2]
        dontusethis =  seriously
        """)))
        settings = {"config": mozconfig}
        plugin = load_from_config("test1", mozconfig)
        config = Configurator(settings=settings)
        config.commit()
        self.failUnless(isinstance(plugin, MetlogPlugin))
        self.failUnless(isinstance(plugin.client, MetlogClient))
        self.failUnless(isinstance(plugin.client.sender, ZmqPubSender))

        client = plugin.client
        sender = client.sender
        bindstrs = sender.pool.socket().connect_bind

        self.assertEquals(bindstrs, \
                ['tcp://localhost:5585', 'tcp://localhost:5586'])

    def test_loading_from_configurator_with_default_sender(self):
        config = get_configurator({})
        client = load_metlog_client(config)
        self.failUnless(isinstance(client.sender, StdLibLoggingSender))

    def test_loading_from_configurator_with_explicit_sender(self):
        config = get_configurator({}, **{
            "metlog.backend": "mozsvc.metrics.MetlogPlugin",
            "metlog.sender_class": "metlog.senders.ZmqPubSender",
            "metlog.sender_bindstrs": "tcp://localhost:5585",
        })
        client = load_metlog_client(config)
        self.failUnless(isinstance(client.sender, ZmqPubSender))
        bindstrs = client.sender.pool.socket().connect_bind
        self.assertEquals(bindstrs, ['tcp://localhost:5585'])


class TestConfigurationLoading(unittest2.TestCase):
    """
    make sure that DecoratorWrapper works on decorators with arguments and with
    out
    """
    def setUp(self):
        if not metlog:
            raise(unittest2.SkipTest('no metlog'))
        mozconfig = Config(StringIO(dedent("""
        [test1]
        backend = mozsvc.metrics.MetlogPlugin
        sender_class=metlog.senders.DebugCaptureSender
        """)))
        settings = {"config": mozconfig}
        self.plugin = load_from_config("test1", mozconfig)
        config = Configurator(settings=settings)
        config.commit()

    def test_loading_from_config(self):
        plugin = self.plugin
        self.assertEqual(len(plugin.client.sender.msgs), 0)

        @timeit
        def target_callable(x, y):
            return x + y

        result = target_callable(5, 6)
        self.assertEqual(result, 11)
        self.assertEqual(len(plugin.client.sender.msgs), 1)

        obj = json.loads(plugin.client.sender.msgs[0])

        expected = 'mozsvc.tests.test_metrics.target_callable'
        actual = obj['fields']['name']
        self.assertEqual(actual, expected)

        # Now test to make sure we can enque 2 messages using stacked
        # decorators
        plugin.client.sender.msgs.clear()
        self.assertEqual(len(plugin.client.sender.msgs), 0)

        @incr_count
        @timeit
        def new_target_callable(x, y):
            return x + y

        result = new_target_callable(5, 6)
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        self.assertEqual(len(msgs), 2)

        # Names should be preserved
        self.assertEqual(new_target_callable.__name__, 'new_target_callable')

        for msg in msgs:
            expected = 'mozsvc.tests.test_metrics.target_callable'
            actual = obj['fields']['name']
            self.assertEqual(actual, expected)

        # First msg should be timer then the counter
        # as decorators just wrap each other
        self.assertEqual(msgs[0]['type'], 'timer')
        self.assertEqual(msgs[1]['type'], 'counter')


class TestCannedDecorators(unittest2.TestCase):
    def setUp(self):
        if not metlog:
            raise(unittest2.SkipTest('no metlog'))
        mozconfig = Config(StringIO(dedent("""
        [test1]
        backend = mozsvc.metrics.MetlogPlugin
        sender_class=metlog.senders.DebugCaptureSender
        """)))
        settings = {"config": mozconfig}
        self.plugin = load_from_config("test1", mozconfig)
        config = Configurator(settings=settings)
        config.commit()

    def test_decorator_ordering(self):
        plugin = self.plugin

        plugin.client.sender.msgs.clear()
        self.assertEqual(len(plugin.client.sender.msgs), 0)

        @incr_count
        @timeit
        def ordering_1(x, y):
            return x + y

        ordering_1(5, 6)
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        self.assertEqual(len(msgs), 2)

        for msg in msgs:
            expected = 'mozsvc.tests.test_metrics.ordering_1'
            actual = msg['fields']['name']
            self.assertEqual(actual, expected)

        # First msg should be counter, then timer as decorators are
        # applied inside to out, but execution is outside -> in
        self.assertEqual(msgs[0]['type'], 'timer')
        self.assertEqual(msgs[1]['type'], 'counter')

        plugin.client.sender.msgs.clear()
        self.assertEqual(len(plugin.client.sender.msgs), 0)

        @timeit
        @incr_count
        def ordering_2(x, y):
            return x + y

        ordering_2(5, 6)
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        self.assertEqual(len(msgs), 2)

        for msg in msgs:
            expected = 'mozsvc.tests.test_metrics.ordering_2'
            actual = msg['fields']['name']
            self.assertEqual(actual, expected)

        # Ordering of log messages should occur in the in->out
        # ordering of decoration
        self.assertEqual(msgs[0]['type'], 'counter')
        self.assertEqual(msgs[1]['type'], 'timer')

    def test_mozsvc_data(self):
        plugin = self.plugin
        plugin.client.sender.msgs.clear()
        msgs = plugin.client.sender.msgs
        self.assertEqual(len(msgs), 0)

        @send_mozsvc_data
        def some_method(request):
            update_mozsvc_data({'foo': 'bar'})

        req = Request({'PATH_INFO': '/foo/bar',
                       'SERVER_NAME': 'somehost.com',
                       'SERVER_PORT': 80,
                       })
        some_method(req)
        msg = json.loads(plugin.client.sender.msgs[0])
        self.assertTrue('foo' in msg['fields'])
        self.assertEqual(msg['fields']['foo'], 'bar')


user_info = MetricsService(name='users', path='/{username}/info',
                           description='some_svc')


@user_info.get(decorators=[timeit, send_mozsvc_data])
def get_info(request):
    return 'foo'


@user_info.get(decorators=[timeit, send_mozsvc_data])
def get_info_mozsvc_data(request):
    update_mozsvc_data({'moe': 'curly'})
    return 'foo'

decorate_all = MetricsService(name='users', path='/{username}/all',
                              description='some_svc')


@decorate_all.get()
def auto_decorate(request):
    update_mozsvc_data({'baz': 'bawlp'})
    return 'foo'


@decorate_all.get(decorators=[incr_count])
def decorator_override(request):
    return 'foo'


class TestMetricsService(unittest2.TestCase):
    def setUp(self):
        if not metlog:
            raise(unittest2.SkipTest('no metlog'))
        mozconfig = Config(StringIO(dedent("""
        [test1]
        backend = mozsvc.metrics.MetlogPlugin
        sender_class=metlog.senders.DebugCaptureSender
        """)))
        settings = {"config": mozconfig}
        self.plugin = load_from_config("test1", mozconfig)
        config = Configurator(settings=settings)
        config.commit()

    def test_metrics_service_get(self):
        req = Request({'PATH_INFO': '/foo/info',
                       'SERVER_NAME': 'somehost.com',
                       'SERVER_PORT': 80,
                       })
        resp = get_info(req)
        self.assertEqual(resp, 'foo')

        plugin = self.plugin
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        self.assertEqual(len(msgs), 1)
        self.assertTrue('timer' in [m['type'] for m in msgs])

    def test_metrics_service_get_mozsvc_data(self):
        req = Request({'PATH_INFO': '/foo/info',
                       'SERVER_NAME': 'somehost.com',
                       'SERVER_PORT': 80,
                       })
        resp = get_info_mozsvc_data(req)
        self.assertEqual(resp, 'foo')

        plugin = self.plugin
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        self.assertEqual(len(msgs), 2)
        self.assertTrue('timer' in [m['type'] for m in msgs])
        self.assertTrue('mozsvc' in [m['type'] for m in msgs])

    def test_decorate_at_constructor(self):
        req = Request({'PATH_INFO': '/foo/all',
                       'SERVER_NAME': 'somehost.com',
                       'SERVER_PORT': 80,
                       })
        resp = auto_decorate(req)
        self.assertEqual(resp, 'foo')

        plugin = self.plugin
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        msg_types = [m['type'] for m in msgs]
        self.assertEqual(len(msgs), 3)
        self.assertTrue('timer' in msg_types)
        self.assertTrue('counter' in msg_types)
        self.assertTrue('mozsvc' in msg_types)

    def test_decorator_override(self):
        req = Request({'PATH_INFO': '/foo/all',
                       'SERVER_NAME': 'somehost.com',
                       'SERVER_PORT': 80,
                       })
        resp = decorator_override(req)
        self.assertTrue(resp == 'foo')

        plugin = self.plugin
        msgs = [json.loads(m) for m in plugin.client.sender.msgs]
        self.assertEqual(len(msgs), 1)
        self.assertTrue('counter' in [m['type'] for m in msgs])



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
import urlparse
import urllib
import time
from decimal import Decimal, InvalidOperation

from pyramid.util import DottedNameResolver


def round_time(value=None, precision=2):
    """Transforms a timestamp into a two digits Decimal.

    Arg:
        value: timestamps representation - float or str.
        If None, uses time.time()

        precision: number of digits to keep. defaults to 2.

    Return:
        A Decimal two-digits instance.
    """
    if value is None:
        value = time.time()
    if not isinstance(value, str):
        value = str(value)
    try:
        digits = '0' * precision
        return Decimal(value).quantize(Decimal('1.' + digits))
    except InvalidOperation:
        raise ValueError(value)


def resolve_name(name, package=None):
    """Resolve dotted name into a python object.

    This function resolves a dotted name as a reference to a python object,
    returning whatever object happens to live at that path.  It's a simple
    convenience wrapper around pyramid's DottedNameResolver.

    The optional argument 'package' specifies the package name for relative
    imports.  If not specified, only absolute paths will be supported.
    """
    return DottedNameResolver(package).resolve(name)


def maybe_resolve_name(name_or_object, package=None):
    """Resolve dotted name or object into a python object.

    This function resolves a dotted name as a reference to a python object,
    returning whatever object happens to live at that path.  If the given
    name is not a string, it is returned unchanged.

    The optional argument 'package' specifies the package name for relative
    imports.  If not specified, only absolute paths will be supported.
    """
    return DottedNameResolver(package).maybe_resolve(name_or_object)


def dnslookup(url):
    """Replaces a hostname by its IP in an url.

    Uses gethostbyname to do a DNS lookup, so the nscd cache is used.

    If gevent has patched the standard library, makes sure it uses the
    original version because gevent uses its own mechanism based on
    the async libevent's evdns_resolve_ipv4, which does not use
    glibc's resolver.
    """
    try:
        from gevent.socket import _socket
        gethostbyname = _socket.gethostbyname
    except ImportError:
        import socket
        gethostbyname = socket.gethostbyname

    # parsing
    parsed_url = urlparse.urlparse(url)
    host, port = urllib.splitport(parsed_url.netloc)
    user, host = urllib.splituser(host)

    # resolving the host
    host = gethostbyname(host)

    # recomposing
    if port is not None:
        host = '%s:%s' % (host, port)

    if user is not None:
        host = '%s@%s' % (user, host)

    parts = [parsed_url[0]] + [host] + list(parsed_url[2:])
    return urlparse.urlunparse(parts)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Simplified API for accessing memcached.

This module provides a simplified API for accessing memcache, modelled on
the API of python-memcached/pylibmc.  It offers useful default behaviours
for serialization, error reporting and connection pooling.
"""

import sys
import time
import json
import traceback
import contextlib
import Queue

import umemcache

import mozsvc
from mozsvc.exceptions import BackendError


class MemcachedClient(object):
    """Helper class for interacting with memcache.

    This class provides the basic methods of the pylibmc Client class, but
    wraps them with some extra functionality:

        * all values are transparently serialized via JSON instead of pickle.
        * connections are taken from an underlying pool.
        * errors are converted into BackendError instances.
        * cas() transparently falls back to add() when appropriate.

    """

    def __init__(self, servers=None, key_prefix="", pool_size=None,
                 pool_timeout=60, logger=None, **kwds):
        if servers is None:
            servers = ["127.0.0.1:11211"]
        elif isinstance(servers, basestring):
            servers = [servers]
        self.key_prefix = key_prefix
        self._logger = logger
        # XXX TODO: umemcache doesn't support clustering.
        # We could implement this ourselves, but is it worth it?
        if len(servers) > 1:
            msg = "Multiple servers are not currently supported. "
            msg += "Consider using moxi for transparent clustering support."
            raise ValueError(msg)
        self.pool = MCClientPool(servers[0], pool_size, pool_timeout)

    @property
    def logger(self):
        """Property to lazily extract a default logger from runtime environ."""
        if self._logger is None:
            try:
                from pyramid.threadlocal import get_current_registry
                self._logger = get_current_registry()["metlog"]
            except (ImportError, KeyError):
                self._logger = mozsvc.logger
        return self._logger

    @contextlib.contextmanager
    def _connect(self):
        """Context mananager for getting a connection to memcached."""
        # We could get an error while trying to create a new connection,
        # or when trying to use an existing connection.  This outer
        # try-except handles the logging for both cases.
        try:
            with self.pool.reserve() as mc:
                # If we get an error while using the client object,
                # disconnect so that it will be removed from the pool.
                try:
                    yield mc
                except (EnvironmentError, RuntimeError), err:
                    if mc is not None:
                        mc.disconnect()
                    raise
        except (EnvironmentError, RuntimeError), err:
            err = traceback.format_exc()
            self.logger.error(err)
            raise BackendError(str(err))

    def get(self, key):
        """Get the value stored under the given key."""
        with self._connect() as mc:
            res = mc.get(self.key_prefix + key)
        if res is None:
            return None
        data, flags = res
        data = json.loads(data)
        return data

    def gets(self, key):
        """Get the current value and casid for the given key."""
        with self._connect() as mc:
            res = mc.gets(self.key_prefix + key)
        if res is None:
            return None, None
        data, flags, casid = res
        data = json.loads(data)
        return data, casid

    def get_multi(self, keys):
        """Get the values stored under the given keys in a single request."""
        with self._connect() as mc:
            prefixed_keys = [self.key_prefix + key for key in keys]
            prefixed_items = mc.get_multi(prefixed_keys)
        items = {}
        for key, res in prefixed_items.iteritems():
            assert key.startswith(self.key_prefix)
            assert res is not None
            data, flags = res
            items[key[len(self.key_prefix):]] = json.loads(data)
        return items

    def set(self, key, value, time=0):
        """Set the value stored under the given key."""
        data = json.dumps(value)
        with self._connect() as mc:
            res = mc.set(self.key_prefix + key, data, time)
        if res != "STORED":
            return False
        return True

    def add(self, key, value, time=0):
        """Add the given key to memcached if not already present."""
        data = json.dumps(value)
        with self._connect() as mc:
            res = mc.add(self.key_prefix + key, data, time)
        if res != "STORED":
            return False
        return True

    def replace(self, key, value, time=0):
        """Replace the given key in memcached if it is already present."""
        data = json.dumps(value)
        with self._connect() as mc:
            res = mc.replace(self.key_prefix + key, data, time)
        if res != "STORED":
            return False
        return True

    def cas(self, key, value, casid, time=0):
        """Set the value stored under the given key if casid matches."""
        data = json.dumps(value)
        with self._connect() as mc:
            # Memcached's CAS only works properly on existing keys.
            # Fortunately ADD has the same semantics for missing keys.
            if casid is None:
                res = mc.add(self.key_prefix + key, data, time)
            else:
                res = mc.cas(self.key_prefix + key, data, casid, time)
        if res != "STORED":
            return False
        return True

    def delete(self, key):
        """Delete the value stored under the given key."""
        with self._connect() as mc:
            res = mc.delete(self.key_prefix + key)
        if res != "DELETED":
            return False
        return True


# Sentinel used to mark an empty slot in the MCClientPool queue.
# Using sys.maxint as the timestamp ensures that empty slots will always
# sort *after* live connection objects in the queue.
EMPTY_SLOT = (sys.maxint, None)


class MCClientPool(object):
    """Pool of umemcache.Client objects, with periodic purging of connections.

    This class implements a simple pool of umemcache Client objects, with
    periodically closing and refreshing of the pooled Client objects.  This
    seems to work around some occasional hangs that were occurring with
    long-lived clients.

    To initialise the pool you must provide the list of server addresses
    to access.  You may also specify the maximum size of the pool and the
    time after which old connections will be recycled.

    To obtain a Client object from the pool, call reserve() as a context
    manager like this::

        with pool.reserve() as mc:
            mc.set("hello", "world")
            assert ms.get("hello") == "world"

    """

    def __init__(self, server, maxsize=None, timeout=60):
        self.server = server
        self.maxsize = maxsize
        self.timeout = timeout
        # Use a synchronized Queue class to hold the active client objects.
        # It will contain tuples (connection_timestamp, client).
        # Using a PriorityQueue ensures that the oldest connection is always
        # used first, allowing them to be closed out when stale.  It also means
        # that a no-maxsize pool can grow and shink according to demand, as old
        # connections are expired and not replaced.
        self.clients = Queue.PriorityQueue(maxsize)
        # If there is a maxsize, prime the queue with empty slots.
        if maxsize is not None:
            for _ in xrange(maxsize):
                self.clients.put(EMPTY_SLOT)

    @contextlib.contextmanager
    def reserve(self):
        """Context-manager to obtain a Client object from the pool."""
        ts, client = self._checkout_client()
        try:
            yield client
        finally:
            self._checkin_client(ts, client)

    def _create_client(self):
        """Create a new Client object."""
        client = umemcache.Client(self.server)
        client.connect()
        return client

    def _checkout_client(self):
        """Checkout a Client ojbect from the pool.

        This method checks out a Client object from the pool, creating a new
        one if necessary.  It will block if a maxsize has been set and there
        are no objects left in the pool
        """
        # If there's no maxsize, no need to block waiting for a connection.
        blocking = (self.maxsize is not None)
        # Loop until we get a non-stale connection, or we create a new one.
        while True:
            try:
                ts, client = self.clients.get(blocking)
            except Queue.Empty:
                # No maxsize and no free connections, create a new one.
                # XXX TODO: we should be using a monotonic clock here.
                now = int(time.time())
                return now, self._create_client()
            else:
                now = int(time.time())
                # If we got an empty slot placeholder, create a new connection.
                if client is None:
                    return now, self._create_client()
                # If the connection is not stale, go ahead and use it.
                if ts + self.timeout > now:
                    return ts, client
                # Otherwise, the connection is stale.
                # Close it, push an empty slot onto the queue, and retry.
                client.disconnect()
                self.clients.put(EMPTY_SLOT)
                continue

    def _checkin_client(self, ts, client):
        """Return a Client object to the pool."""
        # If the connection is now stale, don't return it to the pool.
        # Push an empty slot instead so that it will be refreshed when needed.
        if client.is_connected():
            now = int(time.time())
            if ts + self.timeout > now:
                self.clients.put((ts, client))
            else:
                if self.maxsize is not None:
                    self.clients.put(EMPTY_SLOT)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
import unittest2
import tempfile
import os
import time
import itertools

from mozsvc.secrets import Secrets


class TestSecrets(unittest2.TestCase):

    def setUp(self):
        self._files = []

    def tearDown(self):
        for file in self._files:
            if os.path.exists(file):
                os.remove(file)

    def tempfile(self):
        fd, path = tempfile.mkstemp()
        os.close(fd)
        self._files.append(path)
        return path

    def test_read_write(self):
        secrets = Secrets()

        # We can only add one secret per second to the file, since
        # they are timestamped to 1s resolution.  Fake it.
        real_time = time.time
        time.time = itertools.count(int(real_time())).next
        try:
            secrets.add('phx23456')
            secrets.add('phx23456')
            secrets.add('phx23')
        finally:
            time.time = real_time

        phx23456_secrets = secrets.get('phx23456')
        self.assertEqual(len(secrets.get('phx23456')), 2)
        self.assertEqual(len(secrets.get('phx23')), 1)

        path = self.tempfile()

        secrets.save(path)

        secrets2 = Secrets(path)
        self.assertEqual(len(secrets2.get('phx23456')), 2)
        self.assertEqual(len(secrets2.get('phx23')), 1)
        self.assertEquals(secrets2.get('phx23456'), phx23456_secrets)

    def test_multiple_files(self):
        # creating two distinct files
        secrets = Secrets()
        secrets.add('phx23456')
        one = self.tempfile()
        secrets.save(one)

        secrets = Secrets()
        secrets.add('phx123')
        two = self.tempfile()
        secrets.save(two)

        # loading the two files
        files = one, two
        secrets = Secrets(files)
        keys = secrets.keys()
        keys.sort()
        self.assertEqual(keys, ['phx123', 'phx23456'])



import unittest
from StringIO import StringIO
from textwrap import dedent

from zope.interface import Interface, implements
from zope.interface.verify import verifyObject

from pyramid.config import Configurator, ConfigurationConflictError

from mozsvc.config import Config
from mozsvc.plugin import load_and_register


class ITest1(Interface):
    """A test Interface."""
    pass


class ITest2(Interface):
    """Another test Interface."""
    pass


class Test1(object):
    """A concrete implementation of ITest1."""
    implements(ITest1)

    def __init__(self, **kwds):
        self.kwds = kwds


class Test2(object):
    """A concrete implementation of ITest2."""
    implements(ITest2)

    def __init__(self, **kwds):
        self.kwds = kwds


class Test1And2(object):
    """A concrete implementation of both ITest1 and ITest2."""
    implements(ITest1, ITest2)

    def __init__(self, **kwds):
        self.kwds = kwds


class TestPluginLoading(unittest.TestCase):

    def test_loading_from_config(self):
        config = Config(StringIO(dedent("""
        [test1]
        backend = mozsvc.tests.test_plugin.Test1
        arg1 = 1
        hello = world
        [test2]
        dontusethis =  seriously
        """)))
        settings = {"config": config}
        config = Configurator(settings=settings)
        plugin = load_and_register("test1", config)
        config.commit()
        self.failUnless(verifyObject(ITest1, plugin))
        self.failUnless(isinstance(plugin, Test1))
        self.assertEquals(plugin.kwds["arg1"], 1)
        self.assertEquals(plugin.kwds["hello"], "world")
        self.assertEquals(len(plugin.kwds), 2)
        self.failUnless(config.registry.queryUtility(ITest1) is plugin)

    def test_loading_from_settings(self):
        settings = {
          "test1.backend": "mozsvc.tests.test_plugin.Test1",
          "test1.arg1": 1,
          "test1.hello": "world",
          "test2.dontusethis": "seriously",
        }
        config = Configurator(settings=settings)
        plugin = load_and_register("test1", config)
        config.commit()
        self.failUnless(verifyObject(ITest1, plugin))
        self.failUnless(isinstance(plugin, Test1))
        self.assertEquals(plugin.kwds["arg1"], 1)
        self.assertEquals(plugin.kwds["hello"], "world")
        self.assertEquals(len(plugin.kwds), 2)
        self.failUnless(config.registry.queryUtility(ITest1) is plugin)

    def test_loading_several_plugins(self):
        settings = {
          "test1.backend": "mozsvc.tests.test_plugin.Test1",
          "test1.hello": "world",
          "test2.backend": "mozsvc.tests.test_plugin.Test2",
        }
        config = Configurator(settings=settings)
        plugin1 = load_and_register("test1", config)
        plugin2 = load_and_register("test2", config)
        config.commit()

        self.failUnless(verifyObject(ITest1, plugin1))
        self.failUnless(isinstance(plugin1, Test1))
        self.assertEquals(plugin1.kwds["hello"], "world")
        self.assertEquals(len(plugin1.kwds), 1)
        self.failUnless(config.registry.queryUtility(ITest1) is plugin1)

        self.failUnless(verifyObject(ITest2, plugin2))
        self.failUnless(isinstance(plugin2, Test2))
        self.assertEquals(len(plugin2.kwds), 0)
        self.failUnless(config.registry.queryUtility(ITest2) is plugin2)

    def test_loading_with_conflict_detection(self):
        settings = {
          "test1.backend": "mozsvc.tests.test_plugin.Test1",
          "test_both.backend": "mozsvc.tests.test_plugin.Test1And2",
        }
        config = Configurator(settings=settings)
        load_and_register("test1", config)
        load_and_register("test_both", config)
        self.assertRaises(ConfigurationConflictError, config.commit)

    def test_loading_with_conflict_resolution(self):
        settings = {
          "test1.backend": "mozsvc.tests.test_plugin.Test1",
          "test2.backend": "mozsvc.tests.test_plugin.Test2",
          "test_both.backend": "mozsvc.tests.test_plugin.Test1And2",
        }

        # Load plugin_both last, it will win for both interfaces.
        config = Configurator(settings=settings, autocommit=True)
        load_and_register("test1", config)
        plugin2 = load_and_register("test2", config)
        plugin_both = load_and_register("test_both", config)
        self.failUnless(config.registry.queryUtility(ITest1) is plugin_both)
        self.failUnless(config.registry.queryUtility(ITest2) is plugin_both)

        # Load plugin_both before plugin2, it will be beaten only for that.
        config = Configurator(settings=settings, autocommit=True)
        load_and_register("test1", config)
        plugin_both = load_and_register("test_both", config)
        plugin2 = load_and_register("test2", config)
        self.failUnless(config.registry.queryUtility(ITest1) is plugin_both)
        self.failUnless(config.registry.queryUtility(ITest2) is plugin2)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

import logging
logger = logging.getLogger("mozsvc")


def includeme(config):
    """Include the mozsvc defaults into a Pyramid application config.

    This function allows you to include the following default behaviours
    into your Pyramid application config:

        * add a /__heartbeat__ route and default view implementation

    """
    config.add_route('heartbeat', '/__heartbeat__')
    config.include('mozsvc.tweens')
    config.scan('mozsvc.views')



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****

""" Configuration file reader / writer

https://wiki.mozilla.org/index.php?title=Services/Sync/Server/GlobalConfFile
"""
import re
import os
from ConfigParser import RawConfigParser

from pyramid.config import Configurator

from mozsvc.exceptions import EnvironmentNotFoundError

_IS_NUMBER = re.compile('^-?[0-9].*')
_IS_ENV_VAR = re.compile('\$\{(\w.*)?\}')


def convert(value):
    """Converts a config value"""
    def _get_env(matchobj):
        var = matchobj.groups()[0]
        if var not in os.environ:
            raise EnvironmentNotFoundError(var)
        return os.environ[var]

    def _convert(value):
        if not isinstance(value, basestring):
            # already converted
            return value

        value = value.strip()
        if _IS_NUMBER.match(value):
            try:
                return int(value)
            except ValueError:
                pass
        elif value.startswith('"') and value.endswith('"'):
            return value[1:-1]
        elif value.lower() in ('true', 'false'):
            return value.lower() == 'true'
        return _IS_ENV_VAR.sub(_get_env, value)

    if isinstance(value, basestring) and '\n' in value:
        return [line for line in [_convert(line)
                                  for line in value.split('\n')]
                if line != '']

    return _convert(value)


class Config(RawConfigParser):

    def __init__(self, filename):
        # let's read the file
        RawConfigParser.__init__(self)
        if isinstance(filename, basestring):
            self.filename = filename
            self.read(filename)
        else:
            self.filename = None
            self.readfp(filename)

    def _read(self, fp, filename):
        # first pass
        RawConfigParser._read(self, fp, filename)

        # let's expand it now if needed
        defaults = self.defaults()

        if 'extends' in defaults:
            extends = defaults['extends']
            if not isinstance(extends, list):
                extends = [extends]
            for file_ in extends:
                self._extend(file_)

    def _serialize(self, value):
        """values are serialized on every set"""
        if isinstance(value, bool):
            value = str(value).lower()
        elif isinstance(value, (int, long)):
            value = str(value)
        elif isinstance(value, (list, tuple)):
            value = '\n'.join(['    %s' % line for line in value]).strip()
        else:
            value = str(value)
        return value

    def _unserialize(self, value):
        """values are unserialized on every get"""
        if self.filename is not None:
            here_path = os.path.dirname(self.filename)
            value = value.replace("%(here)s", here_path)
        return convert(value)

    def get_map(self, section=None):
        """returns a dict representing the config set"""
        if section:
            return dict(self.items(section))

        res = {}
        for section in self.sections():
            for option, value in self.items(section):
                option = '%s.%s' % (section, option)
                res[option] = self._unserialize(value)
        return res

    def set(self, section, option, value):
        value = self._serialize(value)
        RawConfigParser.set(self, section, option, value)

    def mget(self, section, option):
        value = self.get(section, option)
        if not isinstance(value, list):
            value = [value]
        return value

    def get(self, section, option):
        value = RawConfigParser.get(self, section, option)
        return self._unserialize(value)

    def items(self, section):
        items = RawConfigParser.items(self, section)
        return [(option, self._unserialize(value)) for option, value in items]

    def _extend(self, filename):
        """Expand the config with another file."""
        if not os.path.isfile(filename):
            raise IOError('No such file: %s' % filename)
        parser = RawConfigParser()
        parser.read([filename])
        here_path = os.path.dirname(filename)
        for section in parser.sections():
            if not self.has_section(section):
                self.add_section(section)
            for option, value in parser.items(section):
                if self.has_option(section, option):
                    continue
                value = value.replace("%(here)s", here_path)
                RawConfigParser.set(self, section, option, value)


def load_into_settings(filename, settings):
    """Load config file contents into a Pyramid settings dict.

    This is a helper function for initialising a Pyramid settings dict from
    a config file.  It flattens the config file sections into dotted settings
    names and updates the given dictionary in place.

    You would typically use this when constructing a Pyramid Configurator
    object, like so::

        def main(global_config, **settings):
            config_file = global_config['__file__']
            load_info_settings(config_file, settings)
            config = Configurator(settings=settings)

    """
    filename = os.path.abspath(
                 os.path.normpath(
                   os.path.expandvars(
                     os.path.expanduser(
                       filename))))

    config = Config(filename)

    # Put values from the config file into the pyramid settings dict.
    for section in config.sections():
        setting_prefix = section.replace(":", ".")
        for name, value in config.get_map(section).iteritems():
            settings[setting_prefix + "." + name] = value

    # Store a reference to the Config object itself for later retrieval.
    settings['config'] = config
    return config


class SettingsDict(dict):
    """A dict subclass with some extra helpers for dealing with app settings.

    This class extends the standard dictionary interface with some extra helper
    methods that are handy when dealing with application settings.  It expects
    the keys to be dotted setting names, where each component indicates one
    section in the settings heirarchy.  You get the following extras:

        * setdefaults:  copy any unset settings from another dict
        * getsection:   return a dict of settings for just one subsection

    """

    separator = "."

    def copy(self):
        """D.copy() -> a shallow copy of D.

        This overrides the default dict.copy method to ensure that the
        copy is also an instance of SettingsDict.
        """
        new_items = self.__class__()
        for k, v in self.iteritems():
            new_items[k] = v
        return new_items

    def getsection(self, section):
        """Get a dict for just one sub-section of the config.

        This method extracts all the keys belonging to the name section and
        returns those values in a dict.  The section name is removed from
        each key.  For example::

            >>> c = SettingsDict({"a.one": 1, "a.two": 2, "b.three": 3})
            >>> c.getsection("a")
            {"one": 1, "two", 2}
            >>>
            >>> c.getsection("b")
            {"three": 3}
            >>>
            >>> c.getsection("c")
            {}

        """
        section_items = self.__class__()
        # If the section is "" then get keys without a section.
        if not section:
            for key, value in self.iteritems():
                if self.separator not in key:
                    section_items[key] = value
        # Otherwise, get keys prefixed with that section name.
        else:
            prefix = section + self.separator
            for key, value in self.iteritems():
                if key.startswith(prefix):
                    section_items[key[len(prefix):]] = value
        return section_items

    def setdefaults(self, *args, **kwds):
        """Import unset keys from another dict.

        This method lets you update the dict using defaults from another
        dict and/or using keyword arguments.  It's like the standard update()
        method except that it doesn't overwrite existing keys.
        """
        for arg in args:
            if hasattr(arg, "keys"):
                for k in arg:
                    self.setdefault(k, arg[k])
            else:
                for k, v in arg:
                    self.setdefault(k, v)
        for k, v in kwds.iteritems():
            self.setdefault(k, v)


def get_configurator(global_config, **settings):
    """Create a pyramid Configurator and populate it with sensible defaults.

    This function is a helper to create and pre-populate a Configurator
    object using the given paste-deploy settings dicts.  It uses the
    mozsvc.config module to flatten the config paste-deploy config file
    into the settings dict so that non-mozsvc pyramid apps can read values
    from it easily.
    """
    # Populate a SettingsDict with settings from the deployment file.
    settings = SettingsDict(settings)
    config_file = global_config.get('__file__')
    if config_file is not None:
        load_into_settings(config_file, settings)
    # Update with default pyramid settings, and then insert for all to use.
    config = Configurator(settings={})
    settings.setdefaults(config.registry.settings)
    config.registry.settings = settings
    return config



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.txt')) as f:
    README = f.read()

with open(os.path.join(here, 'CHANGES.txt')) as f:
    CHANGES = f.read()

requires = ['pyramid', 'simplejson', 'cef']

tests_requires = requires + [
            'pyramid_macauth', 'tokenlib', 'macauthlib>=0.3.0',
            'cornice', 'wsgiproxy', 'unittest2']

extras_require = {
    'metlog': ['metlog-py>=0.9.1'],
    'memcache': ['umemcache>=1.3'],
}


setup(name='mozsvc',
      version='0.6',
      description='Various utilities for Mozilla apps',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Pylons",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        ],
      author='Mozilla Services',
      author_email='services-dev@mozilla.org',
      url='https://github.com/mozilla-services/mozservices',
      keywords='web pyramid pylons',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      extras_require=extras_require,
      tests_require=tests_requires,
      test_suite="mozsvc.tests",
      paster_plugins=['pyramid'])
